--[[
THIS CODE IS TOTALLY UNSUITABLE FOR ANYTHING BUT GETTING A VAGUE SENSE OF THE FINAL PRODUCT

/script Grid2DB = nil; Grid2OptionsDB = nil; ReloadUI()

LibDBLayers-1.0 manages the SavedVariables of your addon.
* Difference between LibDBLayers and AceDB:
** No metatables.  Settings for an object are flattened into a single table.
** Settings can be shared per account (global), class, or spec (custom).
* It offers support for defaults, and a layered hierarchy of settings.
* It is designed to have verbose layered settings in your AddonOptionsDB which is only loaded and used when setting up defaults or changing options.
* The layered settings are flattenned into single compact tables for each object in your AddonDB
* Support for versioning

===== Normal use =====
* Iterate through the character/spec setup and create create your objects and assign their flattened dbx from AddonDB.objects to them.
* Dereference myObject: myObject.dbx.mySavedVariable1 etc.
===== First time use =====
* The lod options of your mod gets loaded
* AddonDB and AddonOptionsDB are created
* Default settings for the current character are added.
===== Upgrade use =====
* An obsolete version causes MyAddonOptions (+ any MyAddonPlugin1Options etc.) to load.
* The new options are added (if possible)

**Note this is still under development and currently guided by Grid2 needs.**
--]]

-- Note this is still under development and currently guided by Grid2 needs.

-- * **profile-current** Index into profile-setup.
-- * **profile-available** For a particular character, the list of (spec specific) profiles profile-current can switch to without loading (lod) MyAddonOptions.
-- * **versions** Stores keys for plugins etc. so they can determine when their settings need updating and thus loading of their lod options.
-- * **profile-setup** Individual setups for all characters/specs in the account.  These are lists of objects found in objects.
-- * **objects** Objects, keyed to a particular layer.  So may contain "heals/account" and / or "heals/druid".
-- * These sections are common to both AddonDB and AddonOptionsDB
-- * The AddonOptionsDB ones are more verbose and separate the various layers.
-- * The AddonDB ones are flattened versions of them.
-- @usage
-- Example AddonDB and AddonOptionsDB (Grid2):
--

local MAJOR, MINOR = "LibDBLayers-1.0", 1
local LibDBLayers, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if not LibDBLayers then return end -- No upgrade needed

local type = type
local pairs, next = pairs, next

LibDBLayers.db_registry = LibDBLayers.db_registry or {}
LibDBLayers.frame = LibDBLayers.frame or CreateFrame("Frame")

local CallbackHandler
local CallbackDummy = { Fire = function() end }


LibDBLayers.defaultSpec = {
	druid = "tree",
	paladin = "holy1",
	priest = "holy2",
	shaman = "resto",
}
local realmKey = GetRealmName()
local charKey = UnitName("player") .. " - " .. realmKey
local _, classKey = UnitClass("player")
classKey = strlower(classKey)
local specKey = LibDBLayers.defaultSpec[classKey]
local scopeChar = "/" -- separator for objectKey/scopeKey
local scopeFormat = "%s/%s" -- format string for objectKey/scopeKey

LibDBLayers.scopes = {"account", classKey, specKey}


function LibDBLayers:GetScopeKey(baseKey, scope, scopes)
	if (scope) then
		return scopeFormat:format(baseKey, scope)
	else
		return scopeFormat:format(baseKey, scopes[# scopes])
	end
end

function LibDBLayers:GetDefaultKeys()
	return charKey, classKey, specKey
end

-- Set up the listed sub tables in the table provided
function LibDBLayers:SetupPath(root, ...)
	local current = root
	for i = 1, select('#', ...), 1 do
		local key = select(i, ...)
		if (not current[key]) then
			current[key] = {}
		end
		current = current[key]
	end
	return current
end

-- Copy (nested) options from src to dst
function LibDBLayers:CopyDB(src, dst)
	if (src) then
		for k, v in pairs(src) do
			if (type(v) == "table") then
				local temp = {}
				self:CopyDB(v, temp)
				dst[k] = temp
			else
				dst[k] = v
			end
		end
	end
end


function LibDBLayers:MakeDefaultDB(db, charKey, classKey, specKey)
	local profileCurrent = self:SetupPath(db, "profile-current")

	local profileCurrentKey = profileCurrent[charKey]
	if (not profileCurrentKey) then
		profileCurrentKey = specKey and (classKey .. "-" .. specKey) or classKey
		profileCurrent[charKey] = profileCurrentKey
	end
	
	local availableList = self:SetupPath(db, "profile-available", charKey)
	if (not availableList[profileCurrentKey]) then
		availableList[profileCurrentKey] = true
	end
		
	local setup = self:SetupPath(db, "profile-setup", profileCurrentKey)
	local objects = self:SetupPath(db, "objects")
	local versions = self:SetupPath(db, "versions")
	
	return setup, objects, versions
end

-- Initialize the mods runtime data.
-- First time through all versions would be missing which subsequently causes options to load and defaults applied.
-- In normal usage settings are used as is and options are not loaded
-- @param modName String the name of the lod Options mod
-- @param svRuntime Table the sv used by the mod at runtime.
-- @param charKeyO String optional UnitName .. "-" .. RealmName.  If not nil, overides the default or current setting.
-- @param classKeyO String optional lowercase of UnitClass.  If not nil, overides the default or current setting.
-- @param specKeyO String optional spec.  If not nil, overides the default or current setting.
-- @usage
-- local dblData = DBL:InitializeRuntime("MyModName", MyModDB)
-- @return dblData mod specific info used as a parameter to most functions
function LibDBLayers:InitializeRuntime(modName, svRuntime, charKeyO, classKeyO, specKeyO)
	local dblData = LibDBLayers.db_registry[modName] or {}
	LibDBLayers.db_registry[modName] = dblData
	
	dblData.svRuntime = svRuntime	-- destination of flattened options
	dblData.charKey = charKeyO or dblData.charKey or charKey
	dblData.classKey = classKeyO or dblData.classKey or classKey
	dblData.specKey = specKeyO or dblData.specKey or specKey
	
	dblData.setupDst, dblData.objectsDst, dblData.versionsDst = LibDBLayers:MakeDefaultDB(svRuntime, dblData.charKey, dblData.classKey, dblData.specKey)

	return dblData
end

-- Initialize the mods options.  These are verbose and layered and lod.
-- @param modName String the name of the lod Options mod
-- @param svRuntime Table the sv used by the mod at runtime.
-- @param charKeyO String optional UnitName .. "-" .. RealmName.  If not nil, overides the default or current setting.
-- @param classKeyO String optional lowercase of UnitClass.  If not nil, overides the default or current setting.
-- @param specKeyO String optional spec.  If not nil, overides the default or current setting.
-- @usage
-- local dblData = DBL:InitializeRuntime("MyModName", MyModOptionsDB)
-- @return dblData mod specific info used as a parameter to most functions
function LibDBLayers:InitializeOptions(modName, svOptions, charKeyO, classKeyO, specKeyO)
	local dblData = LibDBLayers.db_registry[modName] or {}
	LibDBLayers.db_registry[modName] = dblData
	
	dblData.svOptions = svOptions	-- source of layered options that are flattened on their way to svRuntime
	dblData.charKey = charKeyO or dblData.charKey or charKey
	dblData.classKey = classKeyO or dblData.classKey or classKey
	dblData.specKey = specKeyO or dblData.specKey or specKey
	
	dblData.setupSrc, dblData.objectsSrc, dblData.versionsSrc = self:MakeDefaultDB(svOptions, dblData.charKey, dblData.classKey, dblData.specKey)

	return dblData
end


-- Get the data for a particular mod.
-- @param modName String the name of the lod Options mod
-- @usage
-- local dblData = DBL:GetData("MyModName")
-- @return dblData mod specific info used as a parameter to most functions
function LibDBLayers:GetData(modName)
	local dblData = LibDBLayers.db_registry[modName]
	return dblData
end

-- Get the data for a particular mod.
-- @param modName String the name of the lod Options mod
-- @usage
-- local profileCurrent = DBL:GetData("MyModName")
-- @return dblData mod specific info used as a parameter to most functions
function LibDBLayers:GetProfileCurrent(dblData)
	local profileCurrent = dblData.svRuntime["profile-current"][dblData.charKey]
	return profileCurrent
end

-- Loads options if any versions are newer than the one in dblData.versionsDst.
-- @param dblData Table containing dblData.versionsDst, a table of (nested) key/version pairs
-- @param optionsName String the name of the lod Options mod
-- @param funcLoad Function optional Initialization code to call on first load
-- @param ... Pairs of scope, currentVersion. "account" is stored at the top level, other scopes in sub tables.
-- @usage
-- local upgrade = DBL:LoadOptions(versions, optionsName, "account", 1, "druid", 3, "tree", 1)
-- @return upgrade, true if some options need upgrading
function LibDBLayers:LoadOptions(optionsName, dblData, funcLoad, ...)
	local versions, upgrade, forceLoad
	
	if (dblData) then
		versions = dblData.versionsDst
	else
		forceLoad = true
	end

	-- Only load if version is old
	if (versions) then
		local count = select('#', ...)
		assert(count % 2 == 0, "Version parameters need to be multiples of scope, version")
		local checkScope
		for i = 2, count, 2 do
			local scope = select((i - 1), ...)
			local version = select(i, ...)
			if (scope == "account") then
				checkScope = versions
			else
				checkScope = versions[scope]
				if (not checkScope) then
					checkScope = {}
					versions[scope] = checkScope
				end
			end
--print("LibDBLayers:LoadOptions scope:", scope, "version:", version, "=", checkScope[optionsName], checkScope)

			if (not checkScope[optionsName]) then
				checkScope[optionsName] = 0
			end
			if (checkScope[optionsName] < version) then
-- print("LibDBLayers:LoadOptions upgrade needed", optionsName, "current version:", checkScope[optionsName])
				upgrade = true
			end
		end
	end

	if ((upgrade or forceLoad) and not IsAddOnLoaded(optionsName)) then
-- print("LibDBLayers:LoadOptions ==>loading", optionsName)
		LoadAddOn(optionsName)
		if (funcLoad) then
			if (type(funcLoad) == "function") then
				funcLoad()
			elseif (type(funcLoad) == "string") then
				local options = _G[optionsName]
				if (options) then
					options[funcLoad]()
				end
			end
		end
	end
-- print("LibDBLayers:LoadOptions", optionsName, "IsAddOnLoaded:", IsAddOnLoaded(optionsName))
-- print(" ")
	
	return upgrade
end

-- Upgrade defaults if any versions are newer than the one in dblData.versionsDst.
-- @param optionsName String the name of the lod Options mod
-- @param dblData Table containing dblData.versionsDst, a table of (nested) key/version pairs
-- @param funcUpgrade Function Called if the defaults need upgrading
-- @param ... Pairs of scope, currentVersion. "account" is stored at the top level, other scopes in sub tables.
-- @usage
-- local flatten = DBL:UpgradeDefaults(versions, optionsName, MyOptionX.UpgradeDefaults, "account", 1, "druid", 3, "tree", 1)
-- @return flatten, true if some options were upgraded, and thus need to be flattenned and copied to svRuntime
function LibDBLayers:UpgradeDefaults(optionsName, dblData, funcUpgrade, ...)
-- print("LibDBLayers:UpgradeDefaults", optionsName, "funcUpgrade", funcUpgrade, "...", ...)
	local versions = dblData.versionsDst
	local versionsSrc = dblData.versionsSrc
	local flatten

	local count = select('#', ...)
	assert(count % 2 == 0, "Version parameters need to be multiples of scope, version")
	local checkScope
--print("LibDBLayers:UpgradeDefaults", optionsName, "funcUpgrade", funcUpgrade, "count", count)
	for i = 2, count, 2 do
		local scope = select((i - 1), ...)
		local version = select(i, ...)
		if (scope == "account") then
			checkScope = versions
		else
			checkScope = versions[scope]
		end
--print("LibDBLayers:UpgradeDefaults checking", scope, version, "=", checkScope[optionsName], checkScope)

		if (checkScope[optionsName] < version) then
--print("LibDBLayers:UpgradeDefaults checking", scope, version, "=", checkScope[optionsName], checkScope)
			flatten = true

			-- Initialize source versions
			if (scope == "account") then
				checkScope = versionsSrc
			else
				checkScope = versionsSrc[scope]
				if (not checkScope) then
					checkScope = {}
					versionsSrc[scope] = checkScope
				end
			end
			if (not checkScope[optionsName]) then
				checkScope[optionsName] = 0
			end
		end
	end
	
	if (flatten) then
		funcUpgrade(dblData)
	end

	return flatten
end


-- Set up a particular layer 
-- typeList - a table for a particular type in the setup
-- objects - table of object config db indexed by scopeKey
-- scopes - fixed scope layers for the type {"account", <class>, <spec|custom>} etc.
-- scopeIndex - "account", <class>, <spec|custom>
-- db - pairs of scopeKey, config db
function LibDBLayers:SetupLayerObject(typeList, objects, baseKey, scopes, scopeIndex, db)
	local scope = scopes[scopeIndex]
	local scopeKey = scopeFormat:format(baseKey, scope) 
	local layers = typeList[baseKey]
	if (not layers) then
		layers = {}
		typeList[baseKey] = layers
	end
	for i = 1, scopeIndex, 1 do
		layers[i] = scopes[i]	-- {"account", <class>, <spec>} etc.
	end
	if (not objects[scopeKey]) then
		objects[scopeKey] = db
	end
end

function LibDBLayers:SetupMapObject(typeList, objects, scopes, scopeIndex, baseKey, mapKey, mapValue)
	local scope = scopes[scopeIndex]
	local scopeKey = scopeFormat:format(baseKey, scope) 
	local layers = typeList[baseKey]
	if (not layers) then
		layers = {}
		typeList[baseKey] = layers
	end
	for i = 1, scopeIndex, 1 do
		layers[i] = scopes[i]	-- {"account", <class>, <spec>} etc.
	end

	local map = objects[scopeKey]
	if (not map) then
		map = {}
		objects[scopeKey] = map
	end
	
-- print(scopeKey, "->", mapKey, mapValue)
	if (not map[mapKey]) then
		map[mapKey] = mapValue
	end
end

-- For a particular type, apply defaults, and otherwise flatten and copy settings from options to the runtime db 
-- setupSrc - {baseKey1 = {scope1, scope2, ..., scopeN}, ...}
-- objectsSrc - {<baseKey1>/<scope1> = db1, ..., <baseKey1>/<scopeN> = }
-- setupDst - {<baseKey1>/<scopeN>, ...}
-- objectsDst - {<baseKey1>/<scopeN> = {db1 + db2 + ... + dbN}, ...}
function LibDBLayers:FlattenSetupType(setupSrc, objectsSrc, setupDst, objectsDst)
	wipe(setupDst)
	for baseKey, scopes in pairs(setupSrc) do
		-- Copy setup keys
		local scopeKey = scopeFormat:format(baseKey, scopes[# scopes]) 
		setupDst[# setupDst + 1] = scopeKey

		-- Create flattened db
		local dbFlat = objectsDst[scopeKey] or {}
		objectsDst[scopeKey] = dbFlat
		wipe(dbFlat)
		for index, scope in ipairs(scopes) do
			self:CopyDB(objectsSrc[scopeFormat:format(baseKey, scope)], dbFlat)
		end

		-- ToDo: prune unused objects
	end
end

-- Flatten a mapping between two object types (from -> to)
-- Ensure the most scope specific objects are mapped 
-- setupSrc - {baseKey1 = {scope1, scope2, ..., scopeN}, ...}
-- objectsSrc - {<baseKey1>/<scope1> = db1, ..., <baseKey1>/<scopeN> = }
-- setupDst - {<baseKey1>/<scopeN>, ...}
-- objectsDst - {<baseKey1>/<scopeN> = {db1 + db2 + ... + dbN}, ...}
-- type1List - list of scopes for type1
function LibDBLayers:FlattenMap(setupSrc, objectsSrc, setupDst, objectsDst, fromScopes, toScopes)
	wipe(setupDst)
-- print("FlattenMap")
	for baseKey, scopes in pairs(setupSrc) do
		local actualScopes = fromScopes[baseKey]
		local scopeKeyFrom
		
		if (actualScopes) then
			scopeKeyFrom = scopeFormat:format(baseKey, actualScopes[# actualScopes]) 
		else
			-- ToDo upgrade these cases
			scopeKeyFrom = scopeFormat:format(baseKey, "account") 
print("        **old style", scopeKeyFrom)
		end

		-- Copy setup keys
		setupDst[scopeKeyFrom] = true
--print("    scopeKeyFrom", scopeKeyFrom)

		-- Create flattened db
		local dbFlat = objectsDst[scopeKeyFrom] or {}
		objectsDst[scopeKeyFrom] = dbFlat
		wipe(dbFlat)
		for index, scope in ipairs(scopes) do
			local scopeKey = scopeFormat:format(baseKey, scope) 
			local dbSource = objectsSrc[scopeKey]
-- print("    ", scopeKey, dbSource)
			if (dbSource) then
				for mapKeyBase, mapValue in pairs(dbSource) do
					actualScopes = toScopes[mapKeyBase]
					if (actualScopes) then
						local scopeKeyTo = scopeFormat:format(mapKeyBase, actualScopes[# actualScopes])
-- print("        ", scopeKeyFrom, "->", scopeKeyTo, mapValue)
						dbFlat[scopeKeyTo] = mapValue
					else
print("        **old style", scopeKeyFrom, "->", mapKeyBase)
						-- ToDo upgrade these cases
						dbFlat[scopeFormat:format(mapKeyBase, "account")] = mapValue
					end
				end
			end
		end

		-- ToDo: prune unused objects?
	end
-- print(" ")
end

-- Some objects create multiple objects.  This handles the proper naming of these.
function LibDBLayers:GetExtendedKey(scopeKey, extension)
	local extendedKey = gsub(scopeKey, scopeChar, extension .. scopeChar)
	return extendedKey
end

function LibDBLayers:GetBaseKey(scopeKey)
	local baseKey = strsplit(scopeChar, scopeKey)
	return baseKey
end




--[[

local DBObjectLib = {}

--
--	AceDB Utility Functions
--

-- Called to add defaults to a section of the database
--
-- When a ["*"] default section is indexed with a new key, a table is returned
-- and set in the host table.  These tables must be cleaned up by removeDefaults
-- in order to ensure we don't write empty default tables.
local function copyDefaults(dest, src)
	-- this happens if some value in the SV overwrites our default value with a non-table
	--if type(dest) ~= "table" then return end
	for k, v in pairs(src) do
		if k == "*" or k == "**" then
			if type(v) == "table" then
				-- This is a metatable used for table defaults
				local mt = {
					-- This handles the lookup and creation of new subtables
					__index = function(t,k)
							if k == nil then return nil end
							local tbl = {}
							copyDefaults(tbl, v)
							rawset(t, k, tbl)
							return tbl
						end,
				}
				setmetatable(dest, mt)
				-- handle already existing tables in the SV
				for dk, dv in pairs(dest) do
					if not rawget(src, dk) and type(dv) == "table" then
						copyDefaults(dv, v)
					end
				end
			else
				-- Values are not tables, so this is just a simple return
				local mt = {__index = function(t,k) return k~=nil and v or nil end}
				setmetatable(dest, mt)
			end
		elseif type(v) == "table" then
			if not rawget(dest, k) then rawset(dest, k, {}) end
			if type(dest[k]) == "table" then
				copyDefaults(dest[k], v)
				if src['**'] then
					copyDefaults(dest[k], src['**'])
				end
			end
		else
			if rawget(dest, k) == nil then
				rawset(dest, k, v)
			end
		end
	end
end

-- Called to remove all defaults in the default table from the database
local function removeDefaults(db, defaults, blocker)
	for k,v in pairs(defaults) do
		if k == "*" or k == "**" then
			if type(v) == "table" then
				-- Loop through all the actual k,v pairs and remove
				for key, value in pairs(db) do
					if type(value) == "table" then
						-- if the key was not explicitly specified in the defaults table, just strip everything from * and ** tables
						if defaults[key] == nil and (not blocker or blocker[key] == nil) then
							removeDefaults(value, v)
							-- if the table is empty afterwards, remove it
							if next(value) == nil then
								db[key] = nil
							end
						-- if it was specified, only strip ** content, but block values which were set in the key table
						elseif k == "**" then
							removeDefaults(value, v, defaults[key])
						end
					end
				end
			elseif k == "*" then
				-- check for non-table default
				for key, value in pairs(db) do
					if defaults[key] == nil and v == value then
						db[key] = nil
					end
				end
			end
		elseif type(v) == "table" and type(db[k]) == "table" then
			-- if a blocker was set, dive into it, to allow multi-level defaults
			removeDefaults(db[k], v, blocker and blocker[k])
			if next(db[k]) == nil then
				db[k] = nil
			end
		else
			-- check if the current value matches the default, and that its not blocked by another defaults table
			if db[k] == defaults[k] and (not blocker or blocker[k] == nil) then
				db[k] = nil
			end
		end
	end
	-- remove all metatables from the db
	setmetatable(db, nil)
end

-- This is called when a table section is first accessed, to set up the defaults
local function initSection(db, section, svstore, key, defaults)
	local sv = rawget(db, "sv")

	local tableCreated
	if not sv[svstore] then sv[svstore] = {} end
	if not sv[svstore][key] then
		sv[svstore][key] = {}
		tableCreated = true
	end

	local tbl = sv[svstore][key]

	if defaults then
		copyDefaults(tbl, defaults)
	end
	rawset(db, section, tbl)

	return tableCreated, tbl
end

-- Metatable to handle the dynamic creation of sections and copying of sections.
local dbmt = {
	__index = function(t, section)
			local keys = rawget(t, "keys")
			local key = keys[section]
			if key then
				local defaultTbl = rawget(t, "defaults")
				local defaults = defaultTbl and defaultTbl[section]

				if section == "profile" then
					local new = initSection(t, section, "profiles", key, defaults)
					if new then
						-- Callback: OnNewProfile, database, newProfileKey
						t.callbacks:Fire("OnNewProfile", t, key)
					end
				elseif section == "profiles" then
					local sv = rawget(t, "sv")
					if not sv.profiles then sv.profiles = {} end
					rawset(t, "profiles", sv.profiles)
				elseif section == "global" then
					local sv = rawget(t, "sv")
					if not sv.global then sv.global = {} end
					if defaults then
						copyDefaults(sv.global, defaults)
					end
					rawset(t, section, sv.global)
				else
					initSection(t, section, section, key, defaults)
				end
			end

			return rawget(t, section)
		end
}

local function validateDefaults(defaults, keyTbl, offset)
	if not defaults then return end
	offset = offset or 0
	for k in pairs(defaults) do
		if not keyTbl[k] or k == "profiles" then
			error(("Usage: AceDBObject:RegisterDefaults(defaults): '%s' is not a valid datatype."):format(k), 3 + offset)
		end
	end
end

local preserve_keys = {
	["callbacks"] = true,
	["RegisterCallback"] = true,
	["UnregisterCallback"] = true,
	["UnregisterAllCallbacks"] = true,
	["children"] = true,
}

local realmKey = GetRealmName()
local charKey = UnitName("player") .. " - " .. realmKey
local _, classKey = UnitClass("player")
local _, raceKey = UnitRace("player")
local factionKey = UnitFactionGroup("player")
local factionrealmKey = factionKey .. " - " .. realmKey
-- Actual database initialization function
local function initdb(sv, defaults, defaultProfile, olddb, parent)
	-- Generate the database keys for each section

	-- Make a container for profile keys
	if not sv.profileKeys then sv.profileKeys = {} end

	-- map "true" to our "Default" profile
	if defaultProfile == true then defaultProfile = "Default" end

	-- Try to get the profile selected from the char db
	local profileKey = sv.profileKeys[charKey] or defaultProfile or charKey
	sv.profileKeys[charKey] = profileKey

	-- This table contains keys that enable the dynamic creation
	-- of each section of the table.  The 'global' and 'profiles'
	-- have a key of true, since they are handled in a special case
	local keyTbl= {
		["char"] = charKey,
		["realm"] = realmKey,
		["class"] = classKey,
		["race"] = raceKey,
		["faction"] = factionKey,
		["factionrealm"] = factionrealmKey,
		["profile"] = profileKey,
		["global"] = true,
		["profiles"] = true,
	}

	validateDefaults(defaults, keyTbl, 1)

	-- This allows us to use this function to reset an entire database
	-- Clear out the old database
	if olddb then
		for k,v in pairs(olddb) do if not preserve_keys[k] then olddb[k] = nil end end
	end

	-- Give this database the metatable so it initializes dynamically
	local db = setmetatable(olddb or {}, dbmt)

	if not rawget(db, "callbacks") then
		-- try to load CallbackHandler-1.0 if it loaded after our library
		if not CallbackHandler then CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0", true) end
		db.callbacks = CallbackHandler and CallbackHandler:New(db) or CallbackDummy
	end

	-- Copy methods locally into the database object, to avoid hitting
	-- the metatable when calling methods

	if not parent then
		for name, func in pairs(DBObjectLib) do
			db[name] = func
		end
	else
		-- hack this one in
		db.RegisterDefaults = DBObjectLib.RegisterDefaults
		db.ResetProfile = DBObjectLib.ResetProfile
	end

	-- Set some properties in the database object
	db.profiles = sv.profiles
	db.keys = keyTbl
	db.sv = sv
	--db.sv_name = name
	db.defaults = defaults
	db.parent = parent

	-- store the DB in the registry
	AceDB.db_registry[db] = true

	return db
end


--
--	AceDB Object Method Definitions
--

--- Sets the defaults table for the given database object by clearing any
-- that are currently set, and then setting the new defaults.
-- @param defaults A table of defaults for this database
function DBObjectLib:RegisterDefaults(defaults)
	if defaults and type(defaults) ~= "table" then
		error("Usage: AceDBObject:RegisterDefaults(defaults): 'defaults' - table or nil expected.", 2)
	end

	validateDefaults(defaults, self.keys)

	-- Remove any currently set defaults
	if self.defaults then
		for section,key in pairs(self.keys) do
			if self.defaults[section] and rawget(self, section) then
				removeDefaults(self[section], self.defaults[section])
			end
		end
	end

	-- Set the DBObject.defaults table
	self.defaults = defaults

	-- Copy in any defaults, only touching those sections already created
	if defaults then
		for section,key in pairs(self.keys) do
			if defaults[section] and rawget(self, section) then
				copyDefaults(self[section], defaults[section])
			end
		end
	end
end

--- Changes the profile of the database and all of it's namespaces to the
-- supplied named profile
-- @param name The name of the profile to set as the current profile
function DBObjectLib:SetProfile(name)
	if type(name) ~= "string" then
		error("Usage: AceDBObject:SetProfile(name): 'name' - string expected.", 2)
	end

	-- changing to the same profile, dont do anything
	if name == self.keys.profile then return end

	local oldProfile = self.profile
	local defaults = self.defaults and self.defaults.profile

	-- Callback: OnProfileShutdown, database
	self.callbacks:Fire("OnProfileShutdown", self)

	if oldProfile and defaults then
		-- Remove the defaults from the old profile
		removeDefaults(oldProfile, defaults)
	end

	self.profile = nil
	self.keys["profile"] = name
	self.sv.profileKeys[charKey] = name

	-- populate to child namespaces
	if self.children then
		for _, db in pairs(self.children) do
			DBObjectLib.SetProfile(db, name)
		end
	end

	-- Callback: OnProfileChanged, database, newProfileKey
	self.callbacks:Fire("OnProfileChanged", self, name)
end

--- Returns a table with the names of the existing profiles in the database.
-- You can optionally supply a table to re-use for this purpose.
-- @param tbl A table to store the profile names in (optional)
function DBObjectLib:GetProfiles(tbl)
	if tbl and type(tbl) ~= "table" then
		error("Usage: AceDBObject:GetProfiles(tbl): 'tbl' - table or nil expected.", 2)
	end

	-- Clear the container table
	if tbl then
		for k,v in pairs(tbl) do tbl[k] = nil end
	else
		tbl = {}
	end

	local curProfile = self.keys.profile

	local i = 0
	for profileKey in pairs(self.profiles) do
		i = i + 1
		tbl[i] = profileKey
		if curProfile and profileKey == curProfile then curProfile = nil end
	end

	-- Add the current profile, if it hasn't been created yet
	if curProfile then
		i = i + 1
		tbl[i] = curProfile
	end

	return tbl, i
end

--- Returns the current profile name used by the database
function DBObjectLib:GetCurrentProfile()
	return self.keys.profile
end

--- Deletes a named profile.  This profile must not be the active profile.
-- @param name The name of the profile to be deleted
-- @param silent If true, do not raise an error when the profile does not exist
function DBObjectLib:DeleteProfile(name, silent)
	if type(name) ~= "string" then
		error("Usage: AceDBObject:DeleteProfile(name): 'name' - string expected.", 2)
	end

	if self.keys.profile == name then
		error("Cannot delete the active profile in an AceDBObject.", 2)
	end

	if not rawget(self.sv.profiles, name) and not silent then
		error("Cannot delete profile '" .. name .. "'. It does not exist.", 2)
	end

	self.sv.profiles[name] = nil

	-- populate to child namespaces
	if self.children then
		for _, db in pairs(self.children) do
			DBObjectLib.DeleteProfile(db, name, true)
		end
	end

	-- Callback: OnProfileDeleted, database, profileKey
	self.callbacks:Fire("OnProfileDeleted", self, name)
end

--- Copies a named profile into the current profile, overwriting any conflicting
-- settings.
-- @param name The name of the profile to be copied into the current profile
-- @param silent If true, do not raise an error when the profile does not exist
function DBObjectLib:CopyProfile(name, silent)
	if type(name) ~= "string" then
		error("Usage: AceDBObject:CopyProfile(name): 'name' - string expected.", 2)
	end

	if name == self.keys.profile then
		error("Cannot have the same source and destination profiles.", 2)
	end

	if not rawget(self.sv.profiles, name) and not silent then
		error("Cannot copy profile '" .. name .. "'. It does not exist.", 2)
	end

	-- Reset the profile before copying
	DBObjectLib.ResetProfile(self)

	local profile = self.profile
	local source = self.sv.profiles[name]

	copyTable(source, profile)

	-- populate to child namespaces
	if self.children then
		for _, db in pairs(self.children) do
			DBObjectLib.CopyProfile(db, name, true)
		end
	end

	-- Callback: OnProfileCopied, database, sourceProfileKey
	self.callbacks:Fire("OnProfileCopied", self, name)
end

--- Resets the current profile to the default values (if specified).
-- @param noChildren if set to true, the reset will not be populated to the child namespaces of this DB object
function DBObjectLib:ResetProfile(noChildren)
	local profile = self.profile

	for k,v in pairs(profile) do
		profile[k] = nil
	end

	local defaults = self.defaults and self.defaults.profile
	if defaults then
		copyDefaults(profile, defaults)
	end

	-- populate to child namespaces
	if self.children and not noChildren then
		for _, db in pairs(self.children) do
			DBObjectLib.ResetProfile(db)
		end
	end

	-- Callback: OnProfileReset, database
	self.callbacks:Fire("OnProfileReset", self)
end

--- Resets the entire database, using the string defaultProfile as the new default
-- profile.
-- @param defaultProfile The profile name to use as the default
function DBObjectLib:ResetDB(defaultProfile)
	if defaultProfile and type(defaultProfile) ~= "string" then
		error("Usage: AceDBObject:ResetDB(defaultProfile): 'defaultProfile' - string or nil expected.", 2)
	end

	local sv = self.sv
	for k,v in pairs(sv) do
		sv[k] = nil
	end

	local parent = self.parent

	initdb(sv, self.defaults, defaultProfile, self)

	-- fix the child namespaces
	if self.children then
		if not sv.namespaces then sv.namespaces = {} end
		for name, db in pairs(self.children) do
			if not sv.namespaces[name] then sv.namespaces[name] = {} end
			initdb(sv.namespaces[name], db.defaults, self.keys.profile, db, self)
		end
	end

		-- Callback: OnDatabaseReset, database
	self.callbacks:Fire("OnDatabaseReset", self)
	-- Callback: OnProfileChanged, database, profileKey
	self.callbacks:Fire("OnProfileChanged", self, self.keys["profile"])

	return self
end

--- Creates a new database namespace, directly tied to the database.  This
-- is a full scale database in it's own rights other than the fact that
-- it cannot control its profile individually
-- @param name The name of the new namespace
-- @param defaults A table of values to use as defaults
function DBObjectLib:RegisterNamespace(name, defaults)
	if type(name) ~= "string" then
		error("Usage: AceDBObject:RegisterNamespace(name, defaults): 'name' - string expected.", 2)
	end
	if defaults and type(defaults) ~= "table" then
		error("Usage: AceDBObject:RegisterNamespace(name, defaults): 'defaults' - table or nil expected.", 2)
	end
	if self.children and self.children[name] then
		error ("Usage: AceDBObject:RegisterNamespace(name, defaults): 'name' - a namespace with that name already exists.", 2)
	end

	local sv = self.sv
	if not sv.namespaces then sv.namespaces = {} end
	if not sv.namespaces[name] then
		sv.namespaces[name] = {}
	end

	local newDB = initdb(sv.namespaces[name], defaults, self.keys.profile, nil, self)

	if not self.children then self.children = {} end
	self.children[name] = newDB
	return newDB
end

--- Returns an already existing namespace from the database object.
-- @param name The name of the new namespace
-- @param silent if true, the addon is optional, silently return nil if its not found
-- @usage
-- local namespace = self.db:GetNamespace('namespace')
-- @return the namespace object if found
function DBObjectLib:GetNamespace(name, silent)
	if type(name) ~= "string" then
		error("Usage: AceDBObject:GetNamespace(name): 'name' - string expected.", 2)
	end
	if not silent and not (self.children and self.children[name]) then
		error ("Usage: AceDBObject:GetNamespace(name): 'name' - namespace does not exist.", 2)
	end
	if not self.children then self.children = {} end
	return self.children[name]
end

--
--	AceDB Exposed Methods
--

--- Creates a new database object that can be used to handle database settings and profiles.
-- By default, an empty DB is created, using a character specific profile.
--
-- You can override the default profile used by passing any profile name as the third argument,
-- or by passing //true// as the third argument to use a globally shared profile called "Default".
--
-- Note that there is no token replacement in the default profile name, passing a defaultProfile as "char"
-- will use a profile named "char", and not a character-specific profile.
-- @param tbl The name of variable, or table to use for the database
-- @param defaults A table of database defaults
-- @param defaultProfile The name of the default profile. If not set, a character specific profile will be used as the default.
-- You can also pass //true// to use a shared global profile called "Default".
-- @usage
-- -- Create an empty DB using a character-specific default profile.
-- self.db = LibStub("AceDB-3.0"):New("MyAddonDB")
-- @usage
-- -- Create a DB using defaults and using a shared default profile
-- self.db = LibStub("AceDB-3.0"):New("MyAddonDB", defaults, true)
function LibDBLayers:New(tbl, defaults, defaultProfile)
	if type(tbl) == "string" then
		local name = tbl
		tbl = getglobal(name)
		if not tbl then
			tbl = {}
			setglobal(name, tbl)
		end
	end

	if type(tbl) ~= "table" then
		error("Usage: AceDB:New(tbl, defaults, defaultProfile): 'tbl' - table expected.", 2)
	end

	if defaults and type(defaults) ~= "table" then
		error("Usage: AceDB:New(tbl, defaults, defaultProfile): 'defaults' - table expected.", 2)
	end

	if defaultProfile and type(defaultProfile) ~= "string" and defaultProfile ~= true then
		error("Usage: AceDB:New(tbl, defaults, defaultProfile): 'defaultProfile' - string or true expected.", 2)
	end

	return initdb(tbl, defaults, defaultProfile)
end


--]]








--[[

-- @class file
-- @name LibDBHierarchical-1.0.lua
-- @release $Id: LibDBHierarchical-1.0.lua 799 2009-04-09 05:00:05Z Azethoth $
local LibDBHierarchical, oldminor = LibStub:NewLibrary("LibDBHierarchical-1.0", 1)

if not LibDBHierarchical then return end -- No upgrade needed

LibDBHierarchical.db_registry = LibDBHierarchical.db_registry or {}

local type = type
local pairs, next = pairs, next
local rawget, rawset = rawget, rawset
local setmetatable = setmetatable

local realmKey = GetRealmName()
local playerKey = UnitName("player") .. " - " .. realmKey


--{{{ Utility Functions

-- Simple shallow copy for copying defaults
local function copyTable(src, dest)
	if type(dest) ~= "table" then dest = {} end
	if type(src) == "table" then
		for k,v in pairs(src) do
			if type(v) == "table" then
				-- try to index the key first so that the metatable creates the defaults, if set, and use that table
				v = copyTable(v, dest[k])
			end
			dest[k] = v
		end
	end
	return dest
end

-- Called to remove all defaults from partSaved
-- Returns either nil or partSaved containing only properties that are different from partDefault
local function removeDefaults(partSaved, partDefault)
	if (not partSaved) then
		return nil
	end

	for k, v in pairs(partSaved) do
		if (type(v) == "table" and type(partDefault[k]) == "table") then
			partSaved[k] = removeDefaults(v, partDefault[k])
		else
			-- check if the current value matches the default
			if (v == partDefault[k]) then
				partSaved[k] = nil
			end
		end
	end

	-- partSaved is no longer different from partDefault
	if (next(partSaved)) then
		return partSaved
	else
		return nil
	end
end

--}}}


local DBObjectLib = {}

--- Sets the defaults for objectKey for player for all of the parts in its hierarchy.
-- @param objectKey Unique key for a particular object in the hierarchies / levels / parts tables
-- @param player <playerName>-<realmName>
-- @param partKey A key into the parts table
-- @param partDefault A table of defaults for a part
-- @param ... additional pairs of partKey, partDefault in ascending order of priority
function DBObjectLib:UpdateHierarchy(objectKey, player, ...)
	if (not player) then
		player = playerKey
	end
	if (not objectKey) then
		objectKey = "*"
	end

	local hierarchies = self.sv.hierarchies[player]
	if (not hierarchies) then
		hierarchies = {}
		self.sv.hierarchies[player] = hierarchies
	end

	local hierarchy = hierarchies[objectKey]
	if (not hierarchy) then
		hierarchy = {}
		hierarchies[objectKey] = hierarchy
	end

	-- priorities contains interleaved default and saved tables in the order they are processed
	-- higher priority parts are later in the list so their properties can overwrite earlier properties
	local priorities = self.priorities[objectKey]
	if (not priorities) then
		priorities = {}
		self.priorities[objectKey] = priorities
	end

	local partIndex, partKey, partDefault, partSaved
 	local partCount = select('#', ...) / 2
	assert(partCount * 2 == select('#', ...), "part parameters need to be multiples of partKey, partDefault")
 	for i = 1, partCount, 1 do
 		partIndex = i * 2
 		partKey, partDefault = select((partIndex - 1), ...), select((partIndex), ...)

		if (type(partKey) ~= "string") then
			error("Usage: LibDBHierarchical:RegisterHierarchy(player, objectKey, [partKey, partDefault,] ...): 'partKey' - string expected.", 2)
		end
		if (partDefault and type(partDefault) ~= "table") then
			error("Usage: LibDBHierarchical:RegisterHierarchy(player, objectKey, [partKey, partDefault,] ...): 'partDefault' - table or nil expected.", 2)
		end

		-- Update the saved hierarchy
		hierarchy[i] = partKey

		-- Update the runtime list of part priorities
		local part = self.sv.parts[partKey]
		if (part) then
			partSaved = removeDefaults(part[objectKey], partDefault)	-- remove saved attributes no longer different from the default
			part[objectKey] = partSaved
		else
			partSaved = nil
		end
		priorities[partIndex - 1] = partDefault
		priorities[partIndex] = partSaved	-- can be nil which means it contains no overides of the default table
 	end

	-- Trim hierarchy if it got shortened
 	for i = # hierarchy, partCount + 1, -1 do
 		hierarchy[i] = nil
 	end

 	-- Delete hierarchy if it is identical to default
 	local default = hierarchies["*"]
 	if (default and default ~= hierarchy and (# default == # hierarchy)) then
 		local identical = true
		for index, partKey in default do
			if (partKey ~= hierarchy[index]) then
				identical = false
			end
		end
		if (identical) then
			hierarchies[objectKey] = nil
		end
 	end
end


--- Returns a single table containing the collapsed properties from the parts in its hierarchy (both saved and default).
-- @param objectKey Unique key for a particular object in the hierarchies / levels / parts tables
-- @param player <playerName>-<realmName>
function DBObjectLib:GetSettings(objectKey, player)
	if (not player) then
		player = playerKey
	end
	if (not objectKey) then
		error("Usage: LibDBHierarchical:GetSettings(player, objectKey): 'objectKey' - string expected.", 2)
	end
	if (objectKey == "*") then
		error("Usage: LibDBHierarchical:GetSettings(player, objectKey): 'objectKey' - can not be default.", 2)
	end

	local settings = self.settings[objectKey]
	if (not settings) then
		settings = {}
		self.settings[objectKey] = settings
	else
		wipe(settings)
	end

	local priorities = self.priorities[objectKey]
	if (priorities) then
		for index, part in priorities do
			copyTable(part, settings)
		end
	else
		error("Usage: LibDBHierarchical:GetSettings(player, objectKey): 'objectKey' - uninitialized.", 2)
	end

	return settings
end


--- Reset the hierarchy of objectKey for player.
-- @param objectKey Unique key for a particular object in the hierarchies / levels / parts tables
-- @param player <playerName>-<realmName>
function DBObjectLib:ResetHierarchyObject(objectKey, player)
	local priorities = self.priorities[objectKey]
	if (priorities) then
		for index, part in priorities do
			if (part and index % 2  == 0) then
				wipe(part)
			end
		end
	else
		error("Usage: LibDBHierarchical:ResetHierarchyObject(player, objectKey): 'objectKey' - uninitialized.", 2)
	end

	self:GetSettings(objectKey, player)
end


--- Reset the hierarchy of all objects for player.
-- @param player <playerName>-<realmName>
function DBObjectLib:ResetHierarchy(player)
	local priorities = self.priorities[objectKey]
	if (priorities) then
		for index, part in priorities do
			if (part and index % 2  == 0) then
				wipe(part)
			end
		end
	else
		error("Usage: LibDBHierarchical:ResetHierarchyObject(player, objectKey): 'objectKey' - uninitialized.", 2)
	end

	self:GetSettings(objectKey, player)
end



--- Creates a new database object that interleaves and manipulates the levels in a hierarchy and their SV and default components.
--
-- @param tbl The name of variable, or table to use for the database
-- @usage
-- -- Resolve and then ensure that tbl is a table and contains 3 required tables.
-- self.db = LibStub("LibDBHierarchical-1.0"):Initialize("MyAddonSavedVariableName")
function LibDBHierarchical:Initialize(tbl)
	if (type(tbl) == "string") then
		local name = tbl
		tbl = getglobal(name)
		if (not tbl) then
			tbl = {}
			setglobal(name, tbl)
		end
	end

	assert(tbl and type(tbl) == "table", "tbl must be a table")

	if (not tbl.hierarchies) then
		tbl.hierarchies = {}
	end
	if (not tbl.levels) then
		tbl.levels = {}
	end
	if (not tbl.parts) then
		tbl.parts = {}
	end

	local db = {

	}

	for name, func in pairs(DBObjectLib) do
		db[name] = func
	end

	-- Set some properties in the database object
	db.sv = tbl
	db.priorities = {}	-- ascending order list of default and sv tables for a particular objectKey
	db.settings = {}	-- collapsed settings for a particular objectKey

	-- store the DB in the registry
	LibDBHierarchical.db_registry[db] = true

	return db
end



-- upgrade existing databases
for db in pairs(LibDBHierarchical.db_registry) do
	for name, func in pairs(DBObjectLib) do
		db[name] = func
	end
end
--]]